/*
 * (C) 2003 by Chaskiel Grundman <cg2v@andrew.cmu.edu>
 * (C) 2003 by Philip Edelbrock <phil@netroedge.com>
 * Copyright (c) 2008 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Significantly based on the alladin etoken driver:
 * (C) 2002 by Andreas Jellinghaus <aj@dungeon.inka.de>
 * (C) 2002 by Olaf Kirch <okir@caldera.de>
 * 
 * See the file LICENSE for use and distribution information.
 */

/* $Id */

#include "config.h"

#include <usb.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <syslog.h>
#include <errno.h>

#include "egate.h"

#define EGATE_STAT_MASK 0xF0
#define EGATE_STAT_BUSY 0x40
#define EGATE_STAT_RESP 0x20
#define EGATE_STAT_DATA 0x10
#define EGATE_STAT_CMND 0

#define EGATE_CMD_SENDCMD 0x80
#define EGATE_CMD_READ 0x81
#define EGATE_CMD_WRITE    0x82
#define EGATE_CMD_ATR     0x83
#define EGATE_CMD_RESET   0x90
#define EGATE_CMD_FETCHSTATUS  0xa0


#ifdef DEBUG
/* auxiliary functions */
void hexdump(const u_int8_t* buffer, int length)
{
  char line[256];
  int n, i;

  for (i = 0; i < length; i++) {
    if ((i & 0xf) == 0)
      snprintf(line, sizeof(line), "%04x:", i);
    n = strlen(line);
    snprintf(line + n, sizeof(line) - n, " %02hhx", buffer[i]);
    if ((i & 0xf) == 0xf)
      syslog(LOG_DEBUG, "%s", line);
  }
  if (length && (i & 0xf) != 0xf)
    syslog(LOG_DEBUG, "%s", line);
}
#endif


static int do_usb(usb_dev_handle *dev, int requesttype, int request,
                  int value, int index, char *bytes, int size, int timeout)
{
  int rc;

  rc = usb_control_msg(dev, requesttype, request, value, index,
                       bytes, size, timeout);
  if (rc == -1)
    syslog(LOG_ERR, "usb_control_msg returned %u, error is %s",
           rc, usb_strerror());
  return rc;
}


/*
 * Scan for an egate device with the specified USB vendor ID and
 * product ID (normally 0x973/0x001, which are the default if the
 * given values are 0/0).  If filter() is given, it will be
 * called with rock and the dirname and filename of each matching
 * USB device until it returns a nonzero result, indicating a match.
 *
 * After a successful scan, the located device is opened and the
 * handled stored in the egate reader structure; until that reader
 * is released, future calls will return success without doing a
 * new scan, even if different values are given for vendor, product,
 * and filter.
 *
 * Note this currently does _not_ ping the device to see if it is
 * actually there.  Perhaps we should do that in the future.
 *
 * Returns 1 on success, 0 on failure.
 */
int egate_probe(struct egate *egate, int myVendor, int myProduct,
                int (*filter)(void *, char *, char *), void *rock)
{
  char c;
  int rc;
  struct usb_bus *mybus;
  struct usb_device *mydev;

  if (!egate)
    return 0;
  if (!egate->usb) {
    if (myVendor == 0 && myProduct == 0) {
      myVendor  = 0x973;
      myProduct = 0x001;
    }
    usb_init();
    usb_find_busses();
    usb_find_devices();
    for (mybus = usb_busses; mybus; mybus = mybus->next) {
      for (mydev = mybus->devices; mydev; mydev = mydev->next) {
        if ((mydev->descriptor.idVendor == myVendor) &&
            (mydev->descriptor.idProduct == myProduct)) {
          /*
           * Well, this could be it.
           */
          if (filter && !(filter)(rock, mybus->dirname, mydev->filename))
            continue;
          goto found;
        }
      }
    }

    return 0;
found:
    egate->usb = usb_open(mydev);
    if (!egate->usb) {
      syslog(LOG_NOTICE, "%s %d %s: usb_open failed: %s",
             __FILE__, __LINE__, __PRETTY_FUNCTION__,
             usb_strerror());
      return 0;
    }
  }

#if 0
  rc = do_usb(egate->usb, USB_ENDPOINT_IN | USB_TYPE_VENDOR,
              EGATE_CMD_FETCHSTATUS, 0, 0, &c, 1, 10000);
  if (rc != 1)
    return 0;
#endif
  return 1;
}


/* Release the egate device controlled by this reader structure. */
void egate_release(struct egate *egate)
{
  power_down_egate(egate);
  if (egate->usb)
    usb_close(egate->usb);
  egate->usb = NULL;
}


/*
 * Poll the egate device every 100us until it is not busy,
 * then optionally check for the desired status.
 * If there is a USB transfer failure, return -1 (from the caller).
 * If the status result is invalid, return 0 (from the caller).
 */
#define EGATE_POLL(egate, wanted)                                          \
  do {                                                                     \
    do {                                                                   \
      int pollrc;                                                          \
      pollrc = do_usb(egate->usb, USB_ENDPOINT_IN | USB_TYPE_VENDOR,       \
                      EGATE_CMD_FETCHSTATUS, 0, 0,                         \
                      &egate->stat, 1, 10000);                             \
      if (pollrc < 0 && errno == ENODEV)                                   \
        return -1;                                                         \
      if (pollrc != 1)                                                     \
        return 0;                                                          \
    } while ((egate->stat & EGATE_STAT_MASK) == EGATE_STAT_BUSY &&         \
             (usleep(100) || 1));                                          \
    if (wanted != -1 && (egate->stat  & EGATE_STAT_MASK) != wanted) {      \
      syslog(LOG_NOTICE, "Expected state 0x%x, got state 0x%x",            \
             wanted, egate->stat);                                         \
      return 0;                                                            \
    }                                                                      \
  } while(0);


/*
 * Power up and reset the egate card, and fetch its ATR.
 * Returns 1 on success, -1 on USB errors, 0 on other errors.
 */
int power_up_egate(struct egate *egate)
{
  int rc, len;
  u_int8_t buffer[1024];
  struct usb_bus *mybus;
  struct usb_device *mydev;

  egate->atrlen = 0;
  memset(egate->atr, 0, 255);

  /* reset */
  rc = do_usb(egate->usb, USB_ENDPOINT_OUT | USB_TYPE_VENDOR,
              EGATE_CMD_RESET, 0, 0, buffer, 0, 10000);
  if (rc < 0 && errno == ENODEV)
    return -1;
  if (rc < 0)
    return 0;

  EGATE_POLL(egate, EGATE_STAT_CMND);
  /* request atr */
  rc = do_usb(egate->usb, USB_ENDPOINT_IN | USB_TYPE_VENDOR, EGATE_CMD_ATR,
              0, 0, buffer, 255, 10000);
  if (rc < 0 && errno == ENODEV)
    return -1;
  if (rc < 2) {
    /* failed, we should get an atr; minimum atr is TS and T0 */
    return 0;
  }
  len = rc;

  /*
   * otherwise it's not an atr response. inverse convention would be
   * somewhat perverse in a device like this, so I'm not even going to
   * consider it.
   */
  if (buffer[0] != 0x3B)
    return 0;

  /* hey, now we have an egate! */
  egate->atrlen = len;
  memcpy(egate->atr, buffer, len);

  return 1;
}


/*
 * Reset the egate device if it is not in command state.
 * This is called after a failure to insure the device will be in a
 * known state at the start of the next operation.
 *
 * Returns -1 if there is a USB transfer problem, or 0 otherwise;
 * the return code is intended to be returned directly by usb_transfer().
 */
static int egate_condreset(struct egate *egate, int force)
{
  int rc;

  EGATE_POLL(egate, -1);
  if (!force && (egate->stat & EGATE_STAT_MASK) == EGATE_STAT_CMND)
    return 0;
  rc = do_usb(egate->usb, USB_ENDPOINT_OUT | USB_TYPE_VENDOR,
              EGATE_CMD_RESET, 0, 0, 0, 0, 10000);
  if (rc < 0 && errno == ENODEV)
    return -1;
  if (rc < 0)
    return 0;
  EGATE_POLL(egate, -1);
  return 0;
}


/* Reset and power down the egate card, and forget its ATR. */
void power_down_egate(struct egate *egate)
{
  if (egate->usb)
    egate_condreset(egate, 1);
  egate->atrlen = 0;
  memset(egate->atr, 0, 255);
}


/*
 * Do an APDU exchange with the egate card.
 * On success, returns 1 and sets *len_in to the number of bytes received.
 * Returns -1 on USB transfer failure, if we cannot recover by resetting.
 * Returns 0 on other errors.
 */
int usb_transfer(struct egate *egate, u_int8_t *buffer_out,
                 u_int8_t *buffer_in, int len_out, int *len_in)
{
  int rc;
  int read_bytes, write_bytes;
  char tmpbuf[5];

  if (egate == NULL || egate->usb == NULL) {
    /* we need some token */
    syslog(LOG_NOTICE, "token unavailable!");
    return -1;
  }
  EGATE_POLL(egate, EGATE_STAT_CMND);
  if (len_out < 4) {
    syslog(LOG_NOTICE, "Invalid command");
    return 0;
  }
  read_bytes = write_bytes = 0;
  if (len_out == 4) {
    /* Case 1: no Lc or Le */
    /*
     * All T=0 commands nominally transfer data either to or from the
     * card; the data transfer length is never missing (also, data is
     * never transferred in both directions at once; see below).
     * Case 1 commands are treated as transferring 0 bytes to the card;
     * we must provide the 0 length.  Note this looks like masquerading
     * case 1 commands as case 2, but really is not, because the count
     * is a number of bytes _to_ the card (Lc), where zero means zero,
     * and not a number of bytes _from_ the card (Le), where zero means
     * 256.
     */
    len_out = 5;
    memcpy(tmpbuf, buffer_out, 4);
    tmpbuf[4] = 0;
    buffer_out = tmpbuf;
  } else if (len_out > 5) {
    /*
     * Case 3 or 4: P3 is Lc;
     * If the length is 5 + Lc + 1, then the last byte is Le
     */

    write_bytes = len_out - 5;
    if (write_bytes == buffer_out[4] + 1) {
      /*
       * This is a case 4 command, with data going in both directions.
       * The length going to the card (Lc) is in P3 (buffer_out[4]).
       * The length coming from the card (Le) is the last byte in the
       * block (buffer_out[len_out-1]).
       *
       * However, in T=0 data is never transferred in both directions
       * at once, and the egate will barf if we try to send an APDU
       * as case 4.
       *
       * Ideallly, some upper layer should know this and not even try
       * to send case 4 APDU's using T=0.  But the application can't
       * deal, because it doesn't (shouldn't have to) know this is
       * T=0, and the PC/SC framework doesn't have a translation layer,
       * which leaves it (mostly) up to us to deal.
       *
       * We deal by simply omitting the Le byte, sending the APDU as
       * case 3.  The card will handle this correctly and, if the
       * command is successful, will return an 0x61XX status code,
       * indicating to the application that it must fetch the response
       * using GET_RESPONSE, which is a case 2 command.
       */
      read_bytes = buffer_out[len_out-1];
      write_bytes--;
    } else if (write_bytes != buffer_out[4])
      syslog(LOG_ERR, "Buffer length probably incorrect -- help!");
  } else {
    /* Case 2: no Lc, so P3 is Le */
    read_bytes = buffer_out[len_out-1];
    if (read_bytes == 0)
      read_bytes = 256;
  }

#ifdef DEBUG
  syslog(LOG_DEBUG, "Sending command");
  hexdump(buffer_out, 5);
#endif

  /* send via usb */
  rc = do_usb(egate->usb, USB_TYPE_VENDOR, EGATE_CMD_SENDCMD,
              0, 0, buffer_out, 5, 10000);
  if (rc < 0 && errno == ENODEV)
    return -1;
  if (rc != 5) {
    /* not the whole buffer was transmitted ! */
    return egate_condreset(egate, 0);
  }
  EGATE_POLL(egate, -1);

  if (write_bytes > 0) {
    if ((egate->stat & EGATE_STAT_MASK) != EGATE_STAT_DATA) {
#ifdef DEBUG
      syslog(LOG_DEBUG, "-- aborting data transfer");
#endif
    } else {
#ifdef DEBUG
      syslog(LOG_DEBUG, "Sending data");
      hexdump(&buffer_out[5], write_bytes);
#endif

      rc = do_usb(egate->usb, USB_ENDPOINT_OUT | USB_TYPE_VENDOR,
                  EGATE_CMD_WRITE, 0, 0, &buffer_out[5], write_bytes, 10000);
      if (rc < 0 && errno == ENODEV)
        return -1;
      if (rc != write_bytes) {
        /* not the whole buffer was transmitted ! */
        syslog(LOG_NOTICE, "Incomplete data-out transfer");
        return egate_condreset(egate, 0);
      }
      EGATE_POLL(egate, -1);
    }
  }


  if ((egate->stat & EGATE_STAT_MASK) == EGATE_STAT_DATA) {
#ifdef DEBUG
    syslog(LOG_DEBUG, "Reading %d requested bytes of response data",
           read_bytes);
#endif

    /* receive answer via usb */
    rc = do_usb(egate->usb, USB_ENDPOINT_IN | USB_TYPE_VENDOR,
                EGATE_CMD_READ, 0, 0, buffer_in, read_bytes, 10000);
    if (rc < 0 && errno == ENODEV)
      return -1;
    if (rc != read_bytes) {
      syslog(LOG_NOTICE, "Incomplete data-in transfer");
      return egate_condreset(egate, 0);
    }
    EGATE_POLL(egate, EGATE_STAT_RESP);
  } else {
    read_bytes = 0;
    if ((egate->stat & EGATE_STAT_MASK) != EGATE_STAT_RESP) {
      syslog(LOG_NOTICE, "Response not ready; state is 0x%x", egate->stat);
      return egate_condreset(egate, 0);
    }
  }

  /* receive answer via usb */
  rc = do_usb(egate->usb, USB_ENDPOINT_IN | USB_TYPE_VENDOR,
              EGATE_CMD_READ, 0, 0, &buffer_in[read_bytes], 2, 10000);
  if (rc < 0 && errno == ENODEV)
    return -1;
  if (rc < 2) {
    syslog(LOG_NOTICE, "Incomplete response-in transfer");
    /* failed. */
    return egate_condreset(egate, 0);
  }

#ifdef DEBUG
  syslog(LOG_DEBUG, "%s %d %s: received %d, hexdump follows",
         __FILE__, __LINE__, __PRETTY_FUNCTION__, read_bytes +  2);
  hexdump(buffer_in, read_bytes +  2);
#endif

  *len_in = read_bytes + 2;
  return 1;
}
