/*
 * (C) 2003 by Chaskiel Grundman <cg2v@andrew.cmu.edu>
 * (C) 2003 by Philip Edelbrock <phil@netroedge.com>
 * Copyright (c) 2008 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Significantly based on the alladin etoken driver:
 * (C) 2002 by Andreas Jellinghaus <aj@dungeon.inka.de>
 * (C) 2002 by Olaf Kirch <okir@caldera.de>
 * 
 * See the file LICENSE for use and distribution information.
 */

/* $Id */

#include "config.h"

#include <wintypes.h>
#include <pcsclite.h>
#include <ifdhandler.h>
#include <usb.h>
#include <stdio.h>
#include <sys/types.h>
#include <syslog.h>
#include <string.h>

#ifdef HAVE_LIBHAL
#include <libhal.h>
#endif

#include "egate.h"

/*
 * Are we building for v2 or v3 of the IFD handler SPI?
 * Default is to build against v3 if the library supports it.
 */
#if !defined(IFD_HVERSION_3_0) && !defined(IFDHANDLERv2)
#define IFDHANDLERv2 1
#endif

/*
 * IFD_NO_SUCH_DEVICE is new in SPI v3, so the header might
 * not define it, and we shouldn't use it if building for v2.
 */
#if defined(IFDHANDLERv2) && defined(IFD_NO_SUCH_DEVICE)
#undef IFD_NO_SUCH_DEVICE
#define IFD_NO_SUCH_DEVICE IFD_COMMUNICATION_ERROR
#elif !defined(IFD_NO_SUCH_DEVICE)
#define IFD_NO_SUCH_DEVICE IFD_COMMUNICATION_ERROR
#endif


static int egate_init_done = 0;
static struct egate egates[MAX_READER_NUM];


/*
 * Find an egate reader structure for the specified Lun.
 * If none is found, and alloc_if_needed is nonzero, and an
 * empty slot is available, it will be allocated and returned.
 * Otherwise, NULL is returned.
 */
struct egate * get_token_by_lun(DWORD Lun, int alloc_if_needed)
{
  int i, empty = -1;

  if (!egate_init_done) {
    for (i = 0; i < MAX_READER_NUM; i++) {
      memset(egates + i, 0, sizeof(egates[i]));
      egates[i].reader_id = -1;
    }
    egate_init_done = 1;
  }

  /*
   * Find an existing reader matching this Lun,
   * or the first unused reader.
   */
  Lun = (Lun >> 16) & 0xffff;
  for (i = 0; i < MAX_READER_NUM; i++) {
    if (egates[i].reader_id == -1)
      empty = i;
    else if (egates[i].reader_id == Lun)
      return egates + i;
  }

  /* Do they want us to allocate the empty slot? */
  if (!alloc_if_needed)
    return NULL;

  if (empty < 0) {
    /* There are no unused readers; give up */
    return NULL;
  }
  egates[empty].reader_id = Lun;
  return egates + empty;
}


static int egate_usb_filter(void *rock, char *dirname, char *filename)
{
  int i;
  char **names = rock;
  struct usb_device *dev;

  if (names && (strcmp(names[0], dirname) || strcmp(names[1], filename))) {
    /* Not The One */
    return 0;
  }

  for (i = 0; i < MAX_READER_NUM; i++) {
    if (!egates[i].usb)
      continue;
    dev = usb_device(egates[i].usb);
    if (!strcmp(dirname, dev->bus->dirname) &&
        !strcmp(filename, dev->filename))
      return 0;
  }
  return 1;
}


/*** the ifd handler functions ***/


RESPONSECODE IFDHCreateChannel(DWORD Lun, DWORD Channel)
{
  struct egate *egate;

#ifdef DEBUG
  syslog(LOG_DEBUG, "%s %d %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
#endif

  egate = get_token_by_lun(Lun, 1);
  if (! egate)
    return IFD_COMMUNICATION_ERROR;

  if (egate->usb)
    egate_release(egate);

  if (egate_probe(egate, 0, 0, egate_usb_filter, 0))
    return IFD_SUCCESS;
  return IFD_NO_SUCH_DEVICE;
}


#ifdef HAVE_LIBHAL
#ifdef DEBUG
#define lose_plain(x) do {                                               \
  syslog(LOG_DEBUG,                                                      \
         "%s %d %s: %s: error\n",                                        \
         __FILE__, __LINE__, __PRETTY_FUNCTION__, x);                    \
  goto lose;                                                             \
} while (0)
#define lose_dbus(x) do {                                                \
  syslog(LOG_DEBUG,                                                      \
         "%s %d %s: %s: %s: %s\n",                                       \
         __FILE__, __LINE__, __PRETTY_FUNCTION__, x,                     \
         error.name, error.message);                                     \
  goto lose;                                                             \
} while (0)
#else
#define lose_plain(x) goto lose
#define lose_dbus(x)  goto lose
#endif


/* Unfortunately, this is all pretty Linux-specific */
#define prop_bus "usb.bus_number"
#define prop_dev "usb.linux.device_number"
static int lookup_hal_device(char *udi, char **names)
{
  DBusError error;
  DBusConnection *conn = 0;
  LibHalContext *ctx = 0;
  int inited = 1;
  int busno, devno;

  names[0] = names[1] = 0;

  /* set up our HAL connection */
  dbus_error_init(&error);
  if (!(conn = dbus_bus_get(DBUS_BUS_SYSTEM, &error)))
    lose_dbus("dbus_bus_get");
  dbus_connection_set_exit_on_disconnect(conn, 0);
  if (!(ctx = libhal_ctx_new()))
    lose_plain("libhal_ctx_new");
  if (!libhal_ctx_set_dbus_connection(ctx, conn))
    lose_plain("libhal_ctx_set_dbus_connection");
  if (!libhal_ctx_init(ctx, &error))
    lose_dbus("libhal_ctx_init");

  /* get the bus number (dir name) */
  if (!libhal_device_property_exists(ctx, udi, prop_bus, NULL))
    goto lose;
  busno = libhal_device_get_property_int(ctx, udi, prop_bus, &error);
  if (dbus_error_is_set(&error))
    lose_dbus("libhal_device_get_property_string("prop_bus")");
  if (busno < 0 || busno > 999)
    goto lose;
  if (!(names[0] = malloc(4)))
    goto lose;
  sprintf(names[0], "%03d", busno);

  /* get the device number (file name) */
  if (!libhal_device_property_exists(ctx, udi, prop_dev, NULL))
    goto lose;
  devno = libhal_device_get_property_int(ctx, udi, prop_dev, &error);
  if (dbus_error_is_set(&error))
    lose_dbus("libhal_device_get_property_string("prop_dev")");
  if (devno < 0 || devno > 999)
    goto lose;
  if (!(names[1] = malloc(4)))
    goto lose;
  sprintf(names[1], "%03d", devno);

  libhal_ctx_shutdown(ctx, &error);
  libhal_ctx_free(ctx);
  dbus_connection_unref(conn);
  return 1;

lose:
  if (names[0])
    free(names[0]);
  if (names[1])
    free(names[1]);
  if (dbus_error_is_set(&error))
    dbus_error_free(&error);
  if (inited)
    libhal_ctx_shutdown(ctx, &error);
  if (ctx)
    libhal_ctx_free(ctx);
  if (conn)
    dbus_connection_unref(conn);
  return 0;
}
#endif


#ifndef IFDHANDLERv2
RESPONSECODE IFDHCreateChannelByName(DWORD Lun, LPSTR DeviceName)
{
  struct egate *egate;
  int c, vendorid, productid, extrap, l;
  char *names[2] = { 0, 0 }, x = 0, *y;

#ifdef DEBUG
  syslog(LOG_DEBUG, "%s %d %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
#endif

  /*
   * start by parsing the device name, which has the form
   * "usb:<vid>/<pid>" or
   * "usb:<vid>/<pid>:libhal:<udi>" or
   * "usb:<vid>/<pid>:libusb:<dir>:<file>"
   */
  c = sscanf(DeviceName, "usb:%x/%x:%n%c", &vendorid, &productid, &extrap, &x);
  if (c < 2)
    return IFD_NO_SUCH_DEVICE;
  if (x && !strncmp(DeviceName + extrap, "libusb:", 7)) {
    names[0] = DeviceName + extrap + 7;
    names[1] = strchr(names[0], ':');
    if (names[1]) {
      l = names[1] - names[0];
      y = malloc(l + 1);
      if (!y)
        return IFD_COMMUNICATION_ERROR;
      strncmp(y, names[0], l);
      y[l] = 0;
      names[0] = y;

      y = malloc(strlen(names[1] + 1) + 1);
      if (!y) {
        free(names[0]);
        return IFD_COMMUNICATION_ERROR;
      }
      strcpy(y, names[1] + 1);
      names[1] = y;
    } else {
      x = 0;
    }
#ifdef HAVE_LIBHAL
  } else if (x && !strncmp(DeviceName + extrap, "libhal:", 7)) {
    if (!lookup_hal_device(DeviceName + extrap + 7, names))
      x = 0;
#endif
  }

  egate = get_token_by_lun(Lun, 1);
  if (!egate) {
    if (x) {
      free(names[0]);
      free(names[1]);
    }
    return IFD_COMMUNICATION_ERROR;
  }

  if (egate->usb)
    egate_release(egate);

  c = egate_probe(egate, vendorid, productid, egate_usb_filter, names);
  if (x) {
    free(names[0]);
    free(names[1]);
  }
  return c ? IFD_SUCCESS : IFD_NO_SUCH_DEVICE;
}
#endif /* !IFDHANDLERv2 */


RESPONSECODE IFDHCloseChannel(DWORD Lun)
{
  struct egate *egate;

#ifdef DEBUG
  syslog(LOG_DEBUG, "%s %d %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
#endif

  egate = get_token_by_lun(Lun, 0);
  if (! egate || !egate->usb)
    return IFD_NO_SUCH_DEVICE;

  egate_release(egate);
  egate->reader_id = -1;
  return IFD_SUCCESS;
}


RESPONSECODE
IFDHGetCapabilities(DWORD Lun, DWORD Tag, PDWORD Length, PUCHAR Value)
{
  struct egate *egate;

#ifdef DEBUG
  syslog(LOG_DEBUG, "%s %d %s: Lun %ld, Tag %lx\n",
         __FILE__, __LINE__, __PRETTY_FUNCTION__, Lun, Tag);
#endif

  egate = get_token_by_lun(Lun, 0);
  if (! egate || !egate->usb)
    return IFD_NO_SUCH_DEVICE;

  switch (Tag)
  {
    case TAG_IFD_ATR:
      if (*Length > egate->atrlen)
        *Length = egate->atrlen;
      if (*Length)
        memcpy(Value, egate->atr, *Length);
      break;

    case TAG_IFD_SIMULTANEOUS_ACCESS:
      if (*Length >= 1)
      {
        *Length = 1;
        *Value  = MAX_READER_NUM;
        break;
      }

    case TAG_IFD_SLOTS_NUMBER:
      if (*Length >= 1)
      {
        *Length = 1;
        *Value  = 1;
        break;
      }

    default:
      return IFD_ERROR_TAG;
  }
  return IFD_SUCCESS;
}


RESPONSECODE
IFDHSetCapabilities(DWORD Lun, DWORD Tag, DWORD Length, PUCHAR Value)
{
  struct egate *egate;

#if 0 && defined(DEBUG)
  syslog(LOG_DEBUG,
         "%s %d %s: Lun %ld, Tag %lx, Length %ld, hexdump follows\n",
         __FILE__, __LINE__, __PRETTY_FUNCTION__, Lun, Tag, Length);
  hexdump(Value, Length);
#endif

  egate = get_token_by_lun(Lun, 0);
  if (! egate || !egate->usb)
    return IFD_NO_SUCH_DEVICE;

  /* incomplete */
  return IFD_ERROR_TAG;
}


RESPONSECODE
IFDHSetProtocolParameters(DWORD Lun, DWORD Protocol,
                          UCHAR FLAGS, UCHAR PTS1, UCHAR PTS2, UCHAR PTS3)
{
  struct egate *egate;

#ifdef DEBUG
  syslog(LOG_DEBUG, "%s %d %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
#endif

  egate = get_token_by_lun(Lun, 0);
  if (! egate || !egate->usb)
    return IFD_NO_SUCH_DEVICE;

  if (Protocol > 1) {
    /* unsupported protocol */
    return IFD_PROTOCOL_NOT_SUPPORTED;
  }

  return IFD_SUCCESS;
}


RESPONSECODE IFDHPowerICC(DWORD Lun, DWORD Action,
                          PUCHAR Atr, PDWORD AtrLength)
{
  int rc;
  struct egate *egate;

#ifdef DEBUG
  syslog(LOG_DEBUG, "%s %d %s: Lun %ld, Action %ld\n",
         __FILE__, __LINE__, __PRETTY_FUNCTION__, Lun, Action);
#endif

  egate = get_token_by_lun(Lun, 0);
  if (! egate || !egate->usb)
    return IFD_NO_SUCH_DEVICE;

  *AtrLength = 0;

  switch (Action) {
    case IFD_RESET:
      power_down_egate(egate);
      /* fall through */

    case IFD_POWER_UP:
      if (power_up_egate(egate) > 0) {
        *AtrLength = egate->atrlen;
        memcpy(Atr, egate->atr, *AtrLength);
        return IFD_SUCCESS;
      }
#ifdef DEBUG
      syslog(LOG_DEBUG, "%s %d %s: power up/reset failed\n",
             __FILE__, __LINE__, __PRETTY_FUNCTION__);
#endif
      /*
       * Failure might not indicate the USB device is gone,
       * but certainly indicates a problem that is unlikely
       * to go away with another try.
       */
      return IFD_NO_SUCH_DEVICE;

    case IFD_POWER_DOWN:
      power_down_egate(egate);
      return IFD_SUCCESS;

    default:
      return IFD_NOT_SUPPORTED;
  }
}


#ifdef IFDHANDLERv2
RESPONSECODE IFDHControl(DWORD Lun, PUCHAR TxBuffer, DWORD TxLength,
                         PUCHAR RxBuffer, PDWORD pRxLength)
#else
RESPONSECODE IFDHControl(DWORD Lun, DWORD dwControlCode, PUCHAR TxBuffer,
                         DWORD TxLength, PUCHAR RxBuffer, DWORD RxLength,
                         LPDWORD pdwBytesReturned)
#endif
{
#ifdef IFDHANDLERv2
  DWORD RxLength = *pRxLength;
#endif
  struct egate *egate;

#ifdef DEBUG
  syslog(LOG_DEBUG,
         "%s %d %s: Lun %ld TxLength %ld RxLength %ld, hexdump follows\n",
         __FILE__, __LINE__, __PRETTY_FUNCTION__,
         Lun, TxLength, RxLength);
  hexdump(TxBuffer, TxLength);
#endif

#ifdef IFDHANDLERv2
  *pRxLength = 0;
#else
  *pdwBytesReturned = 0;
#endif

  egate = get_token_by_lun(Lun, 0);
  if (! egate || !egate->usb)
    return IFD_NO_SUCH_DEVICE;

  /* Control is for reader communication.
   * Implementation is optional. */
  return IFD_COMMUNICATION_ERROR;
}


RESPONSECODE IFDHICCPresence(DWORD Lun)
{
  struct egate *egate;

#if 0 && defined(DEBUG)
  syslog(LOG_DEBUG, "%s %d %s: Lun %ld\n",
         __FILE__, __LINE__, __PRETTY_FUNCTION__, Lun);
#endif

  egate = get_token_by_lun(Lun, 0);
  if (! egate || !egate->usb)
    return IFD_NO_SUCH_DEVICE;

#if 0
  /*
   * If we want to do this, we need to make egate_probe
   * actually do such a probe, or else add an egate_ping.
   */
  if (!egate_probe(egate))
    return IFD_ICC_NOT_PRESENT;
#endif

  return IFD_ICC_PRESENT;
}


RESPONSECODE IFDHTransmitToICC(DWORD Lun,
                               SCARD_IO_HEADER SendPci,
                               PUCHAR TxBuffer,
                               DWORD TxLength,
                               PUCHAR RxBuffer,
                               PDWORD RxLength, PSCARD_IO_HEADER RecvPci)
{
  struct egate *egate;
  int rc, len_in;

#ifdef DEBUG
  syslog(LOG_DEBUG,
         "%s %d %s: Lun %ld Protocol %ld TxLength %ld RxLength %ld\n",
         __FILE__, __LINE__, __PRETTY_FUNCTION__, Lun,
         SendPci.Protocol, TxLength, *RxLength);
#endif

  len_in = *RxLength;
  *RxLength = 0; /* should be 0 on all errors */

  egate = get_token_by_lun(Lun, 0);
  if (! egate || !egate->usb)
    return IFD_NO_SUCH_DEVICE;

  if (SendPci.Protocol > 1) {
    /* only T=0 supported */
    /* I don't know how T=1 should differ for this device,
       since alot of ISO7816-3 (and all of 1 and 2) doesn't
       apply to it; the transfer protocol is usb and doesn't
       resemble either T=0 (ACK bytes) or T=1 (CRC) */
    /* the existing devices don't have a TDi value in their ATR,
       so we should only expect to see T=0 requests anyway */
    return IFD_PROTOCOL_NOT_SUPPORTED;
  }

  if ((TxBuffer == NULL) || (RxBuffer == NULL) || (RxLength == NULL)) {
    /* these should not be NULL */
    return IFD_COMMUNICATION_ERROR;
  }

  rc = usb_transfer(egate, TxBuffer, RxBuffer, TxLength, &len_in);

#ifdef DEBUG
  syslog(LOG_DEBUG, "%s %d %s: Status %d, RxLength %ld\n",
         __FILE__, __LINE__, __PRETTY_FUNCTION__, rc, *RxLength);
#endif
  if (rc > 0) {
    *RxLength = len_in;
    return IFD_SUCCESS;
  }
  if (rc < 0)
    return IFD_NO_SUCH_DEVICE;
  return IFD_COMMUNICATION_ERROR;
}
